from .base import APIObject, StrEnum


class InputKeyValue(StrEnum):
    """Values of remote key names."""

    STANDBY = "Standby"
    BACK = "Back"
    FIND = "Find"
    RED = "RedColour"
    GREEN = "GreenColour"
    YELLOW = "YellowColour"
    BLUE = "BlueColour"
    HOME = "Home"
    VOLUME_UP = "VolumeUp"
    VOLUME_DOWN = "VolumeDown"
    MUTE = "Mute"
    OPTIONS = "Options"
    DOT = "Dot"
    DIGIT_0 = "Digit0"
    DIGIT_1 = "Digit1"
    DIGIT_2 = "Digit2"
    DIGIT_3 = "Digit3"
    DIGIT_4 = "Digit4"
    DIGIT_5 = "Digit5"
    DIGIT_6 = "Digit6"
    DIGIT_7 = "Digit7"
    DIGIT_8 = "Digit8"
    DIGIT_9 = "Digit9"
    INFO = "Info"
    CURSOR_UP = "CursorUp"
    CURSOR_DOWN = "CursorDown"
    CURSOR_LEFT = "CursorLeft"
    CURSOR_RIGHT = "CursorRight"
    CONFIRM = "Confirm"
    NEXT = "Next"
    PREVIOUS = "Previous"
    ADJUST = "Adjust"
    WATCH_TV = "WatchTV"
    VIEWMODE = "Viewmode"
    TELETEXT = "Teletext"
    SUBTITLE = "Subtitle"
    CHANNEL_STEP_UP = "ChannelStepUp"
    CHANNEL_STEP_DOWN = "ChannelStepDown"
    SOURCE = "Source"
    AMBILIGHT_ON_OFF = "AmbilightOnOff"
    PLAY_PAUSE = "PlayPause"
    PAUSE = "Pause"
    FAST_FORWARD = "FastForward"
    STOP = "Stop"
    REWIND = "Rewind"
    RECORD = "Record"
    ONLINE = "Online"


class InputKey(APIObject):
    """Model of remote key input request."""

    key: InputKeyValue
    """Value of the key."""



from .ambilight import (
    AmbilightColor,
    AmbilightColors,
    AmbilightColorSettings,
    AmbilightLayer,
    AmbilightMode,
    AmbilightModeValue,
    AmbilightPower,
    AmbilightPowerValue,
    AmbilightTopology,
)
from .applications import (
    Application,
    ApplicationComponent,
    ApplicationIntent,
    Applications,
    ApplicationShort,
)
from .audio import CurrentVolume, Volume
from .base import APIObject, StrEnum, ValidationError
from .channels import (
    AllChannels,
    Channel,
    ChannelID,
    ChannelList,
    ChannelListID,
    ChannelShort,
    CurrentChannel,
    SetChannel,
)
from .general import PowerState, PowerStateValue
from .input import InputKey, InputKeyValue
from .pairing import (
    DeviceInfo,
    PairingAuthInfo,
    PairingGrantPayload,
    PairingRequestPayload,
    PairingRequestResponse,
    PairingResponse,
)

__all__ = [
    "APIObject",
    "AllChannels",
    "AmbilightColor",
    "AmbilightColorSettings",
    "AmbilightColors",
    "AmbilightLayer",
    "AmbilightMode",
    "AmbilightModeValue",
    "AmbilightPower",
    "AmbilightPowerValue",
    "AmbilightTopology",
    "Application",
    "ApplicationComponent",
    "ApplicationIntent",
    "ApplicationShort",
    "Applications",
    "Channel",
    "ChannelID",
    "ChannelList",
    "ChannelListID",
    "ChannelShort",
    "CurrentChannel",
    "CurrentVolume",
    "DeviceInfo",
    "InputKey",
    "InputKeyValue",
    "PairingAuthInfo",
    "PairingGrantPayload",
    "PairingRequestPayload",
    "PairingRequestResponse",
    "PairingResponse",
    "PowerState",
    "PowerStateValue",
    "SetChannel",
    "StrEnum",
    "ValidationError",
    "Volume",
]



from .base import APIObject, StrEnum


class PowerStateValue(StrEnum):
    """TV power state values."""

    ON = "On"
    STANDBY = "Standby"


class PowerState(APIObject):
    """Model of a power state request and response."""

    powerstate: PowerStateValue
    """A power state value."""



from .base import APIObject


class Volume(APIObject):
    """Model of a volume set request."""

    current: int
    """Volume value."""
    muted: bool = False
    """Whether the TV is muted."""


class CurrentVolume(Volume):
    """Model of a volume response."""

    min: int
    """Minimum allowed volume value."""
    max: int
    """Maximum allowed volume value."""



from typing import List

from pydantic.fields import Field

from .base import APIObject


class ApplicationComponent(APIObject):
    """Model of application component API object."""

    package_name: str = Field(alias="packageName")
    """Application package name."""
    class_name: str = Field(alias="className")
    """Application class name."""


class ApplicationIntent(APIObject):
    """Model of application intent API object."""

    component: ApplicationComponent
    """Application component."""
    action: str
    """Action within application."""


class ApplicationShort(APIObject):
    """Model of an activity launch request body."""

    intent: ApplicationIntent
    """Definition of application intent."""


class Application(ApplicationShort):
    """Model of an application in application list response body."""

    label: str
    """Application name."""
    order: int
    """Application order."""
    id: str
    """Application ID."""
    type: str
    """Application type."""


class Applications(APIObject):
    """Model of applications list response body."""

    version: int
    """List version."""
    applications: List[Application]
    """List of application definitions."""



from typing import List, Optional

from .base import APIObject


class DeviceInfo(APIObject):
    """Model of information about device requesting pairing."""

    id: str
    """ID of the device."""
    device_name: str
    """Name of the device."""
    device_os: str
    """OS of the device."""
    app_id: str
    """ID of the application."""
    app_name: str
    """Name of the application."""
    type: str
    """Type of the application."""


class PairingAuthInfo(APIObject):
    """Model of authentication information in pairing confirmation request."""

    pin: str
    """Value of the PIN displayed on the TV."""
    auth_timestamp: int
    """Timestamp value from response to the pairing request."""
    auth_signature: str
    """HMAC signature created from the PIN value."""


class PairingRequestPayload(APIObject):
    """Model of a pairing request."""

    scope: List[str]
    """Requested permissions scope. Available options are `read`, `write`, `control`."""
    device: DeviceInfo
    """Requesting device info."""


class PairingGrantPayload(APIObject):
    """Model of pairing confirmation request."""

    auth: PairingAuthInfo
    """Authentication info."""
    device: DeviceInfo
    """Requesting device info."""


class PairingResponse(APIObject):
    """Model of response the pairing request or confirmation request.

    Note:
        Even tough there's "error" in the field names it doesn't mean an actual error occured.
        The TV communicates successful states using those fields as well.

    """

    error_id: str
    """ID of the pairing process state."""
    error_text: str
    """Description of the pairing process state."""


class PairingRequestResponse(PairingResponse):
    """Model of a response to the pairing request.

    Note:
        Even tough there's "error" in the field names it doesn't mean an actual error occured.
        The TV communicates successful states using those fields as well.

    """

    auth_key: Optional[str] = None
    """Authentication key."""
    timestamp: Optional[int] = None
    """Timestamp sent by TV."""
    timeout: Optional[int] = None
    """Timeout before the pairing is cancelled."""



from .api import PhilipsTVAPI
from .exceptions import (
    PhilipsError,
    PhilipsTVAPIError,
    PhilipsTVAPIMalformedResponseError,
    PhilipsTVAPIUnauthorizedError,
    PhilipsTVError,
    PhilipsTVPairingError,
    PhilipsTVRemoteError,
)
from .model import DeviceInfo
from .pairing import PhilipsTVPairer
from .remote import AmbilightColor, InputKeyValue, PhilipsTVRemote
from .tv import PhilipsTV

__version__ = "0.4.0"  # This will be set during build by poetry-dynamic-versioning

__all__ = [
    "AmbilightColor",
    "DeviceInfo",
    "InputKeyValue",
    "PhilipsError",
    "PhilipsTV",
    "PhilipsTVAPI",
    "PhilipsTVAPIError",
    "PhilipsTVAPIMalformedResponseError",
    "PhilipsTVAPIUnauthorizedError",
    "PhilipsTVError",
    "PhilipsTVPairer",
    "PhilipsTVPairingError",
    "PhilipsTVRemote",
    "PhilipsTVRemoteError",
    "__version__",
]



import sys


def wrapped_cli() -> None:
    try:
        from ._cli import cli
    except ModuleNotFoundError:
        print("CLI dependencies missing!", file=sys.stderr, flush=True)
        print("Please install the package as 'philipstv[cli]'.", file=sys.stderr, flush=True)
        sys.exit(1)
    else:
        cli()


if __name__ == "__main__":
    wrapped_cli()



from enum import Enum
from typing import Any, Type, TypeVar

from pydantic import BaseModel
from pydantic import ValidationError as ValidationError  # noqa: F401 | intentional reexport

_SelfAPIObject = TypeVar("_SelfAPIObject", bound="APIObject")


class APIObject(BaseModel):
    """Base for all API objects.

    Note:
        In all :class:`APIObject` subclassess, all `attributes` are also `parameters` accepted in
        a constructor.

    """

    def dump(self) -> Any:
        """Dump the object's JSON data.

        This JSON data (usually a dict) represents a body of some API request or response.

        Returns:
            Model's JSON data.

        """
        data = super().dict(by_alias=True)
        return data["__root__"] if self.__custom_root_type__ else data

    @classmethod
    def parse(cls: Type[_SelfAPIObject], raw: Any) -> _SelfAPIObject:
        """Construct the API object from given JSON data.

        Each :class:`APIObject` can be created from JSON data (usually a dict) of some API request
        or response body.

        Args:
            raw: Request or response body JSON data.

        Returns:
            An instance of a subclass this is called on.

        Raises:
            ValidationError: if given JSON data cannot be parsed into this model.

        """
        return cls.parse_obj(raw)

    class Config:
        allow_population_by_field_name = True
        use_enum_values = True


class StrEnum(str, Enum):
    """Enum with string values."""

    pass



from typing import Dict, Tuple, Union

from pydantic.fields import Field

from .base import APIObject, StrEnum


class AmbilightPowerValue(StrEnum):
    """Ambilight power state values."""

    ON = "On"
    OFF = "Off"


class AmbilightPower(APIObject):
    """Ambilight power state request and response model."""

    power: AmbilightPowerValue
    """Ambilight power state value."""


class AmbilightTopology(APIObject):
    """Ambilight topology response model."""

    layers: int
    """Number of layers."""
    left: int
    """Number of pixels on the left edge."""
    top: int
    """Number of pixels on the top edge."""
    right: int
    """Number of pixels on the right edge."""
    bottom: int
    """Number of pixels on the bottom edge."""


class AmbilightModeValue(StrEnum):
    """Ambilight mode values."""

    INTERNAL = "internal"
    MANUAL = "manual"
    EXPERT = "expert"


class AmbilightMode(APIObject):
    """Ambilight mode request and response model."""

    current: AmbilightModeValue
    """Ambiglight mode value."""


class AmbilightColor(APIObject):
    """Ambilight color model."""

    r: int = Field(ge=0, le=255)
    """Red color component."""
    g: int = Field(ge=0, le=255)
    """Green color component."""
    b: int = Field(ge=0, le=255)
    """Blue color component."""

    @classmethod
    def from_tuple(cls, color_tuple: Tuple[int, int, int]) -> "AmbilightColor":
        """Create :class:`AmbilightColor` instance from a tuple.

        Args:
            color_tuple: Tuple of three integers representing RGB color.

        Returns:
            Model instance.

        """
        return cls(r=color_tuple[0], g=color_tuple[1], b=color_tuple[2])


class AmbilightLayer(APIObject):
    """Model of a single Ambilight layer per-pixel colors definition.

    Each edge is a mappping of a pixel name to it's color::

        layer = AmbilightLayer(top={"0": AmbilightColor(...), "1": AmbilightColor(...)})

    """

    left: Dict[str, AmbilightColor] = Field(default_factory=dict)
    """Colors of pixels on the left edge."""
    top: Dict[str, AmbilightColor] = Field(default_factory=dict)
    """Colors of pixels on the top edge."""
    right: Dict[str, AmbilightColor] = Field(default_factory=dict)
    """Colors of pixels on the right edge."""
    bottom: Dict[str, AmbilightColor] = Field(default_factory=dict)
    """Colors of pixels on the bottom edge."""


class AmbilightColors(APIObject):
    """Model of full Amblight per-pixel color definition.

    This model is actually a wrapper around a dict since layer names are not predefined.
    Values should be set as dict using ``__root__`` kwarg in constructor and accessed as in normal
    dict using indexing::

        colors = AmbilightColors(__root__={"layer1": AmbilightLayer()})
        layer1 = colors["layer1"]

    """

    __root__: Dict[str, AmbilightLayer] = {}
    """Mapping of layer name to :class:`AmbilightLayer`."""

    def __getitem__(self, item: str) -> AmbilightLayer:
        return self.__root__[item]


AmbilightColorSettings = Union[
    AmbilightColor,  # single color for all pixels
    AmbilightColors,  # individual pixels
]



import json
import logging
from pathlib import Path
from typing import Any, Dict, Optional, Tuple

from appdirs import user_data_dir
from pydantic import BaseModel, BaseSettings, ValidationError
from pydantic.env_settings import SettingsSourceCallable

_LOGGER = logging.getLogger(__name__)


DATA_FILE = Path(user_data_dir("philipstv", "cyran.dev")) / "data.json"


def json_data_source(settings: BaseSettings) -> Dict[str, Any]:
    try:
        _LOGGER.debug("Trying to load application data from %s", DATA_FILE)
        return json.loads(DATA_FILE.read_text())  # type: ignore [no-any-return]
    except FileNotFoundError:
        _LOGGER.debug("Data file not found")
        return {}
    finally:
        _LOGGER.debug("Application data loaded successfully")


class HostData(BaseModel):
    host: str
    id: str
    key: str


class PhilipsTVData(BaseSettings):
    last_host: HostData

    @classmethod
    def load(cls) -> Optional["PhilipsTVData"]:
        try:
            return cls()
        except ValidationError:
            return None

    def save(self) -> None:
        _LOGGER.debug("Saving application data to %s", DATA_FILE)
        if not DATA_FILE.exists():
            _LOGGER.debug("Data file doesn't exist, creating")
            DATA_FILE.parent.mkdir(parents=True, exist_ok=True)
            DATA_FILE.touch()
        DATA_FILE.write_text(self.json())
        _LOGGER.debug("Application data saved successfully")

    class Config:
        @classmethod
        def customise_sources(
            cls,
            init_settings: SettingsSourceCallable,
            env_settings: SettingsSourceCallable,
            file_secret_settings: SettingsSourceCallable,
        ) -> Tuple[SettingsSourceCallable, ...]:
            return (init_settings, json_data_source)



import logging
from base64 import b64decode
from typing import Callable

from ._utils import create_signature
from .api import PhilipsTVAPI
from .exceptions import PhilipsTVPairingError
from .model import DeviceInfo, PairingAuthInfo, PairingGrantPayload, PairingRequestPayload
from .types import Credentials

LOGGER = logging.getLogger(__name__)

_SECRET = b64decode(
    "JCqdN5AcnAHgJYseUn7ER5k3qgtemfUvMRghQpTfTZq7Cvv8EPQPqfz6dDxPQPSu4gKFPWkJGw32zyASgJkHwCjU"
)


PinCallback = Callable[[], str]


class PhilipsTVPairer:
    """Encapsulates the pairing process and allows to perform it in a single call."""

    def __init__(self, api: PhilipsTVAPI, device_info: DeviceInfo) -> None:
        """
        Args:
            api: :class:`PhilipsTVAPI` instance with which pairing will be performed.
            device_info: Informations about the device which requests the pairing. This doesn't
                really seem to have any effect, besides the fact that `id` given here serves as a
                "username" and later will be the first value in the credentials tuple.
        """
        self._api: PhilipsTVAPI = api
        self.device_info = device_info

    def pair(self, pin_callback: PinCallback) -> Credentials:
        """Perfom the pairing.

        Hint:
            If the pairing is successful, this leaves the underlying :class:`PhilipsTVAPI` instance
            in authenticated state, so you don't have to set :attr:`PhilipsTVAPI.auth` with the
            received credentials.

        Args:
            pin_callback: Callback function which should return the PIN displayed on the TV
                screen as a string. It's impossible to know the PIN before requesting pairing
                so this function should be used to prompt the user for the PIN.

        Returns:
            Authentication credentials tuple. This value can be used to authenticate with the TV
            at any point in future.

        """
        self._api.auth = None
        pair_response = self._api.pair_request(self._get_request_payload())

        is_success = pair_response.error_id == "SUCCESS"
        if not (is_success and pair_response.auth_key and pair_response.timestamp):
            raise PhilipsTVPairingError(pair_response)

        pair_pin = pin_callback()

        self._api.auth = (self.device_info.id, pair_response.auth_key)
        grant_response = self._api.pair_grant(
            self._get_grant_payload(pair_pin, pair_response.timestamp)
        )

        if grant_response.error_id != "SUCCESS":
            self._api.auth = None
            raise PhilipsTVPairingError(grant_response)

        return (self.device_info.id, pair_response.auth_key)

    def _get_request_payload(self) -> PairingRequestPayload:
        return PairingRequestPayload(
            scope=["read", "write", "control"],
            device=self.device_info,
        )

    def _get_grant_payload(self, pin: str, timestamp: int) -> PairingGrantPayload:
        signature = create_signature(_SECRET, f"{timestamp}{pin}".encode()).decode()
        return PairingGrantPayload(
            auth=PairingAuthInfo(pin=pin, auth_timestamp=timestamp, auth_signature=signature),
            device=self.device_info,
        )



import platform
from typing import Dict, List, Optional, Union

from ._utils import create_device_id
from .api import PhilipsTVAPI
from .exceptions import PhilipsTVRemoteError
from .model import (
    AmbilightColor,
    AmbilightColors,
    AmbilightLayer,
    AmbilightPower,
    AmbilightPowerValue,
    AmbilightTopology,
    Application,
    Channel,
    ChannelID,
    DeviceInfo,
    InputKey,
    InputKeyValue,
    PowerState,
    PowerStateValue,
    SetChannel,
    Volume,
)
from .pairing import PhilipsTVPairer, PinCallback
from .tv import PhilipsTV
from .types import Credentials

__all__ = ["AmbilightColor", "InputKeyValue", "PhilipsTVRemote"]


class PhilipsTVRemote:
    """High level interface to the TV.

    Wraps the API into convenient methods simulating functions that could be found on a TV remote.

    Can be used either with existing :class:`PhilipsTVAPI` instance, or create it itself.
    In the following example ``remote_one`` and ``remote_two`` are equivalent::

        api = PhilipsTVAPI(PhilipsTV("192.169.0.100", auth=("id", "key")))
        remote_one = PhilipsTVRemote(api)
        remote_two = PhilipsTVRemote.new("192.168.0.100", ("id", "key"))

    """

    def __init__(self, api: PhilipsTVAPI) -> None:
        """
        Args:
            api: Instance of an API to be used by the remote.

        """
        self._api = api
        self._channels_cache: List[Channel] = []
        self._applications_cache: List[Application] = []
        self._ambilight_topology_cache: Optional[AmbilightTopology] = None

    @property
    def host(self) -> str:
        """IP address of the underlying :class:`PhilipsTVAPI` instance.

        Warning:
            This value is read-only.

        """
        return self._api.host

    @property
    def auth(self) -> Optional[Credentials]:
        """Credentials used for authentication in the underlying :class:`PhilipsTVAPI` instance.

        Hint:
            This value can be set and changed at any moment during :class:`PhilipsTVRemote` usage.

        """
        return self._api.auth

    @auth.setter
    def auth(self, value: Optional[Credentials]) -> None:
        self._api.auth = value

    @classmethod
    def new(cls, host: str, auth: Optional[Credentials] = None) -> "PhilipsTVRemote":
        """Create a new remote for given host without the need to inject:class:`PhilipsTVAPI`
        instance.

        Args:
            host: IP address of the TV.
            auth: Authentication credentials. If not given, the only feature you will be able to
                use is pairing: :func:`pair`.

        """
        return cls(PhilipsTVAPI(PhilipsTV(host=host, auth=auth)))

    def pair(self, pin_callback: PinCallback, id: Optional[str] = None) -> Credentials:
        """Perform pairing with the TV.

        After successful pairing this remote instance will be authenticated, even if you didn't
        provide credentials when creating it.

        Args:
            pin_callback: A function taking no arguments and returning the PIN displayed
                on TV as a string.
            id: Device ID to use for pairing. This will be later used as a first value in
                credentials tuple. If not given, a random, 16 characters, alphanumeric string will
                be used.

        Returns:
            A tuple of acquired authentication credentials.

        """
        id = id or create_device_id()
        uname_info = platform.uname()
        device_info = DeviceInfo(
            id=id,
            device_name=uname_info.node,
            device_os=uname_info.system,
            app_id=69,
            app_name="philipstv",
            type="native",
        )
        pairer = PhilipsTVPairer(self._api, device_info)
        return pairer.pair(pin_callback)

    def get_power(self) -> bool:
        """Return the current power state.

        Returns:
            A power state. `True` means on, `False` means standby.

        """
        return True if self._api.get_powerstate().powerstate == PowerStateValue.ON else False

    def set_power(self, power: bool) -> None:
        """Set current power state.

        Args:
            power: Power state to set. `True` means on, `False` means standby.

        """
        value = PowerStateValue.ON if power is True else PowerStateValue.STANDBY
        self._api.set_powerstate(PowerState(powerstate=value))

    def get_volume(self) -> int:
        """Return current volume."""
        return self._api.get_volume().current

    def set_volume(self, volume: int) -> None:
        """Set current volume.

        Args:
            volume: Volume value to set.

        """
        self._api.set_volume(Volume(current=volume))

    def get_current_channel(self) -> str:
        """Return current TV channel.

        If the TV currently displays something different than television (any app), this will return
        the last watched channel.

        Returns:
            Current TV channel name.

        """
        return self._api.get_current_channel().channel.name

    def set_channel(self, channel: Union[int, str]) -> None:
        """Change to the given TV channel.

        Use :func:`get_all_channels` to find valid values.

        If the TV currently displays something different than television (any app), it will go to
        the television.

        Args:
            channel: Number or name of the channel to change to.

        Raises:
            PhilipsTVRemoteError: If invalid channel number or name is given.

        """
        if not self._channels_cache:
            self._channels_cache = self._api.get_all_channels().channel

        if isinstance(channel, str):
            matching = filter(lambda chan: chan.name == channel, self._channels_cache)
        elif isinstance(channel, int):
            matching = filter(lambda chan: chan.preset == str(channel), self._channels_cache)

        found_channel = next(matching, None)
        if not found_channel:
            raise PhilipsTVRemoteError(f"Channel '{channel}' not available")

        self._api.set_channel(SetChannel(channel=ChannelID(ccid=found_channel.ccid)))

    def get_all_channels(self) -> Dict[int, str]:
        """Return all available channels and their numbers.

        Returns:
            A mapping of channel number to channel name.

        """
        self._channels_cache = self._api.get_all_channels().channel
        return {int(channel.preset): channel.name for channel in self._channels_cache}

    def input_key(self, key: InputKeyValue) -> None:
        """Emulate pressing a key on the TV remote.

        Args:
            key: A key value to send to the TV.

        """
        self._api.input_key(InputKey(key=key))

    def get_ambilight_power(self) -> bool:
        """Return current ambilight power state.

        Returns:
            Ambilight power state. `True` means on, `False` means off.

        """
        return True if self._api.get_ambilight_power().power == AmbilightPowerValue.ON else False

    def set_ambilight_power(self, power: bool) -> None:
        """Set ambilight power state.

        Args:
            power: Ambilight power state to set. `True` means on, `False` means off.

        """
        value = AmbilightPowerValue.ON if power is True else AmbilightPowerValue.OFF
        self._api.set_ambilight_power(AmbilightPower(power=value))

    def set_ambilight_color(
        self,
        color: Optional[AmbilightColor] = None,
        *,
        left: Optional[AmbilightColor] = None,
        top: Optional[AmbilightColor] = None,
        right: Optional[AmbilightColor] = None,
        bottom: Optional[AmbilightColor] = None,
    ) -> None:
        """Set ambilight color.

        On sides without given color, the color will not be changed. ``color`` defines color for all
        sides. If any of ``left``, ``top``, ``right`` or ``bottom`` args are given, they override
        ``color`` on that side.

        Args:
            color: A color to set on all sides.
            left: A color to set on the left side.
            top: A color to set on the top side.
            right: A color to set on the right side.
            bottom: A color to set on the bottom side.

        """
        if color and not any((left, top, right, bottom)):
            self._api.set_ambilight_cached(color)
            return

        if not self._ambilight_topology_cache:
            self._ambilight_topology_cache = self._api.get_ambilight_topology()

        topology = self._ambilight_topology_cache
        sides = {}
        if set_left := (left or color):
            sides["left"] = self._create_ambilight_side(set_left, topology.left)
        if set_top := (top or color):
            sides["top"] = self._create_ambilight_side(set_top, topology.top)
        if set_right := (right or color):
            sides["right"] = self._create_ambilight_side(set_right, topology.right)
        if set_bottom := (bottom or color):
            sides["bottom"] = self._create_ambilight_side(set_bottom, topology.bottom)

        colors = AmbilightColors(__root__={"layer1": AmbilightLayer(**sides)})
        self._api.set_ambilight_cached(colors)

    def _create_ambilight_side(
        self, color: AmbilightColor, points: int
    ) -> Dict[str, AmbilightColor]:
        return {str(point): color for point in range(points)}

    def get_applications(self) -> List[str]:
        """Return a list of available applications.

        Returns:
            List of application names.

        """
        self._applications_cache = self._api.get_applications().applications
        return [app.label for app in self._applications_cache]

    def launch_application(self, application: str) -> None:
        """Launch an application.

        Use :func:`get_applications` to find valid values.

        Args:
            application: An application name.

        Raises:
            PhilipsTVRemoteError: If invalid application name is given.

        """
        if not self._applications_cache:
            self._applications_cache = self._api.get_applications().applications

        matching = filter(lambda app: app.label == application, self._applications_cache)

        found_application = next(matching, None)
        if not found_application:
            raise PhilipsTVRemoteError(f"Application '{application}' not available")

        self._api.launch_application(found_application)



import logging
import time
from dataclasses import dataclass
from functools import wraps
from typing import Any, Callable, Optional, Tuple, Union

import click
from click.exceptions import UsageError

from philipstv import __version__

from ._data import HostData, PhilipsTVData
from .exceptions import (
    PhilipsError,
    PhilipsTVAPIUnauthorizedError,
    PhilipsTVError,
    PhilipsTVPairingError,
    PhilipsTVRemoteError,
)
from .model import AmbilightColor, InputKeyValue
from .remote import PhilipsTVRemote

_LOGGER = logging.getLogger(__name__)

ColorArg = Tuple[int, int, int]


COMMON_ERROR_MESSAGE = """\
You can try to run the application with '--debug' option to get more info."""

UNAUTHORIZED_MESSAGE = f"""\
Could not authenticate with the TV.
Make sure you use correct credentials.
{COMMON_ERROR_MESSAGE}"""

CONNECTION_ERROR_MESSAGE = f"""\
Could not connect with the TV.
Make sure the IP address is correct and TV is powered on.
{COMMON_ERROR_MESSAGE}"""

UNKNOWN_ERROR_MESSAGE = f"""\
Unknown error occured.
{COMMON_ERROR_MESSAGE}"""


def handle_tv_errors(func: Callable[..., None]) -> Callable[..., None]:
    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        try:
            return func(*args, **kwargs)
        except PhilipsTVRemoteError as err:
            _LOGGER.debug("Remote error", exc_info=True)
            raise click.ClickException(str(err))
        except PhilipsTVAPIUnauthorizedError:
            _LOGGER.debug("Authentication error", exc_info=True)
            raise click.ClickException(UNAUTHORIZED_MESSAGE)
        except PhilipsTVError:
            _LOGGER.debug("Connection error", exc_info=True)
            raise click.ClickException(CONNECTION_ERROR_MESSAGE)
        except PhilipsError:
            _LOGGER.debug("Unknown error", exc_info=True)
            raise click.ClickException(UNKNOWN_ERROR_MESSAGE)

    return wrapper


KEY_MAP = {
    "standby": InputKeyValue.STANDBY,
    "back": InputKeyValue.BACK,
    "find": InputKeyValue.FIND,
    "red": InputKeyValue.RED,
    "green": InputKeyValue.GREEN,
    "yellow": InputKeyValue.YELLOW,
    "blue": InputKeyValue.BLUE,
    "home": InputKeyValue.HOME,
    "volup": InputKeyValue.VOLUME_UP,
    "voldown": InputKeyValue.VOLUME_DOWN,
    "mute": InputKeyValue.MUTE,
    "options": InputKeyValue.OPTIONS,
    "dot": InputKeyValue.DOT,
    "0": InputKeyValue.DIGIT_0,
    "1": InputKeyValue.DIGIT_1,
    "2": InputKeyValue.DIGIT_2,
    "3": InputKeyValue.DIGIT_3,
    "4": InputKeyValue.DIGIT_4,
    "5": InputKeyValue.DIGIT_5,
    "6": InputKeyValue.DIGIT_6,
    "7": InputKeyValue.DIGIT_7,
    "8": InputKeyValue.DIGIT_8,
    "9": InputKeyValue.DIGIT_9,
    "info": InputKeyValue.INFO,
    "up": InputKeyValue.CURSOR_UP,
    "down": InputKeyValue.CURSOR_DOWN,
    "left": InputKeyValue.CURSOR_LEFT,
    "right": InputKeyValue.CURSOR_RIGHT,
    "ok": InputKeyValue.CONFIRM,
    "next": InputKeyValue.NEXT,
    "prev": InputKeyValue.PREVIOUS,
    "adjust": InputKeyValue.ADJUST,
    "tv": InputKeyValue.WATCH_TV,
    "view": InputKeyValue.VIEWMODE,
    "teletext": InputKeyValue.TELETEXT,
    "subtitle": InputKeyValue.SUBTITLE,
    "chanup": InputKeyValue.CHANNEL_STEP_UP,
    "chandown": InputKeyValue.CHANNEL_STEP_DOWN,
    "source": InputKeyValue.SOURCE,
    "ambilight": InputKeyValue.AMBILIGHT_ON_OFF,
    "play": InputKeyValue.PLAY_PAUSE,
    "pause": InputKeyValue.PAUSE,
    "forward": InputKeyValue.FAST_FORWARD,
    "stop": InputKeyValue.STOP,
    "rewind": InputKeyValue.REWIND,
    "rec": InputKeyValue.RECORD,
    "online": InputKeyValue.ONLINE,
}


@dataclass
class TVContext:
    remote: PhilipsTVRemote
    host: Optional[str] = None
    id: Optional[str] = None
    key: Optional[str] = None
    save: bool = False


pass_tv_context = click.make_pass_decorator(TVContext)

CONTEXT_SETTINGS = {"help_option_names": ["-h", "--help"]}


@click.group(context_settings=CONTEXT_SETTINGS)
@click.option("-a", "--host", type=click.STRING, help="TV IP address.")
@click.option("-i", "--id", type=click.STRING, help="Connecting device ID.")
@click.option("-k", "--key", type=click.STRING, help="Connecting device secret key.")
@click.option(
    "-s", "--save", is_flag=True, default=False, help="Save host, ID and key for future use."
)
@click.help_option("-h", "--help")
@click.version_option(__version__, "-v", "--version")
@click.option("-d", "--debug", is_flag=True, default=False, help="Enable debug log.")
@click.pass_context
@handle_tv_errors
def cli(
    ctx: click.Context,
    host: Optional[str],
    id: Optional[str],
    key: Optional[str],
    save: bool,
    debug: bool,
) -> None:
    """Welcome to philipstv - a CLI remote control for Philips Android-powered TVs.

    Before you will be able to use this program, you need to pair it with your TV. For this,
    ensure your TV is powered on. The simplest way to perform pairing is by running:

    \b
        philipstv --host IP --save pair

    To learn more about pairing process run:

    \b
        philipstv pair -h

    In fact, you can learn more about every command by running it, followed by '-h'.
    The whole tool can be explored this way.

    Enjoy!
    """
    log_level = logging.DEBUG if debug else logging.INFO
    logging.basicConfig(level=log_level)

    # we need to treat pairing exceptionally
    if ctx.invoked_subcommand == "pair":
        # I'm passing empty string to `PhilipsTVRemote` if no host is given...
        # This is pretty fucking bad, but I really don't want to make 'remote' field optional
        # just for this one command
        # 'pair' HAS TO VALIDATE the 'host' before doing anything!
        ctx.obj = TVContext(PhilipsTVRemote.new(host or ""), host, id, key, save)
        return

    # for all other use cases we need either all data given or none given but saved data available
    if not (host and id and key):  # if not all auth is given
        if host or id or key:  # but some is given
            raise click.UsageError("Options --host, --id and --key have to be used together.")

        # at this point we know there's no auth given at all
        if save:  # so let's exit if user tries to save non-existent data
            raise click.UsageError("--save requires giving --host, --id and --key.")

        if not (saved_data := PhilipsTVData.load()):  # let's exit if nothing is saved as well
            raise click.UsageError("No TV data (--host, --id, --key) given or saved.")

        # something was saved!
        host = saved_data.last_host.host
        id = saved_data.last_host.id
        key = saved_data.last_host.key
        _LOGGER.debug("Using saved data: %r", saved_data.last_host)
    elif save:  # if all auth is given and we want to save
        PhilipsTVData(last_host=HostData(host=host, id=id, key=key)).save()

    ctx.obj = TVContext(PhilipsTVRemote.new(host, (id, key)), host, id, key, save)


@cli.command("pair")
@pass_tv_context
@click.pass_context
@handle_tv_errors
def pair(ctx: click.Context, tv_ctx: TVContext) -> None:
    """Pair with the TV to obtain authentication credentials.

    In order to perform pairing, you have to provide the TV IP address (--host option).
    You can find it in the TV's network settings. Look for something like "Show network
    configuration".

    Additionally, you can also provide device ID (--id), which will be later used during
    authentication. If device ID is not provided, random 16 characters long alphanumeric string
    will be generated.

    During pairing you will be prompted to enter the PIN number displayed on the TV screen.

    After successful pairing, your credentials will be displayed. If you don't want to enter them
    every time you run any command, you can pass (--save) option. This will save the credentials
    in a file. Saved credentials will be automatically used in future, unless other credentials are
    explicitly provided.

    \b
    Example valid pairing commands:
        philipstv --host IP pair
        philipstv --host IP --id ID pair
        philipstv --host IP --id ID --save pair

    """
    if not tv_ctx.host:
        raise click.UsageError("No host given (--host).")
    if tv_ctx.key:
        raise click.UsageError("Option --key is invalid in pairing context.")

    def prompt_pin() -> str:
        return str(click.prompt(text="Enter PIN displayed on the TV", type=click.STRING))

    try:
        credentials = tv_ctx.remote.pair(prompt_pin, tv_ctx.id)
    except PhilipsTVPairingError as err:
        raise click.ClickException(str(err))

    click.echo("Pairing successful!")
    click.echo(f"ID:\t{credentials[0]}")
    click.echo(f"Key:\t{credentials[1]}")

    if tv_ctx.save:
        PhilipsTVData(
            last_host=HostData(host=tv_ctx.host, id=credentials[0], key=credentials[1])
        ).save()
        click.echo("Credentials saved.")


@cli.group("power", help="Manage power state.")
def power() -> None:
    pass


@power.command("get", help="Get current power state.")
@pass_tv_context
@handle_tv_errors
def power_get(tv_ctx: TVContext) -> None:
    click.echo("on" if tv_ctx.remote.get_power() else "off")


@power.command("set", help="Set power state.")
@click.argument("power", type=click.Choice(("on", "off")))
@pass_tv_context
@handle_tv_errors
def power_set(tv_ctx: TVContext, power: str) -> None:
    tv_ctx.remote.set_power(True if power == "on" else False)


@cli.group("volume", help="Manage audio volume.")
def volume() -> None:
    pass


@volume.command("get", help="Get current audio volume.")
@pass_tv_context
@handle_tv_errors
def volume_get(tv_ctx: TVContext) -> None:
    click.echo(tv_ctx.remote.get_volume())


@volume.command("set", help="Set audio volume.")
@click.argument("volume", type=click.INT)
@pass_tv_context
@handle_tv_errors
def volume_set(tv_ctx: TVContext, volume: int) -> None:
    tv_ctx.remote.set_volume(volume)


@cli.group("channel", help="Manage TV channels.")
def channel() -> None:
    pass


@channel.command("get", help="Get current TV channel.")
@pass_tv_context
@handle_tv_errors
def channel_get(tv_ctx: TVContext) -> None:
    click.echo(tv_ctx.remote.get_current_channel())


@channel.command("list", help="List all available TV channels.")
@pass_tv_context
@handle_tv_errors
def channel_list(tv_ctx: TVContext) -> None:
    click.echo("\n".join(f"{no}\t{chan}" for no, chan in tv_ctx.remote.get_all_channels().items()))


@channel.command("set", help="Set TV channel.")
@click.argument("channel", type=click.STRING)
@pass_tv_context
@handle_tv_errors
def channel_set(tv_ctx: TVContext, channel: str) -> None:
    set_channel: Union[str, int] = channel
    if channel.isdigit():
        set_channel = int(set_channel)
    tv_ctx.remote.set_channel(set_channel)


@cli.command("key")
@click.argument(
    "keys", type=click.Choice(tuple(KEY_MAP), case_sensitive=False), nargs=-1, required=True
)
@click.option(
    "--delay", "-d", type=int, default=0, help="Delay (in milliseconds) between consecutive keys."
)
@pass_tv_context
@handle_tv_errors
def key(tv_ctx: TVContext, keys: Tuple[str, ...], delay: Optional[int]) -> None:
    """Emulate pressing keys on the TV remote.

    You can provide any number of key names, separated by a space. They will be sent to the TV in
    the given order.

    When providing multiple keys, they can be sent faster than the TV can react. To prevent this,
    you can use '--delay' option.
    """
    tv_ctx.remote.input_key(KEY_MAP[keys[0]])
    for key in keys[1:]:
        if delay:
            time.sleep(delay / 1000)
        tv_ctx.remote.input_key(KEY_MAP[key])


@cli.group("ambilight", help="Manage ambilight.")
def ambilight() -> None:
    pass


@ambilight.group("power", help="Manage ambilight power.")
def ambilight_power() -> None:
    pass


@ambilight_power.command("get", help="Get current ambilight power state.")
@pass_tv_context
@handle_tv_errors
def ambilight_power_get(tv_ctx: TVContext) -> None:
    click.echo("on" if tv_ctx.remote.get_ambilight_power() else "off")


@ambilight_power.command("set", help="Set ambilight power state.")
@click.argument("power", type=click.Choice(("on", "off")))
@pass_tv_context
@handle_tv_errors
def ambilight_power_set(tv_ctx: TVContext, power: str) -> None:
    tv_ctx.remote.set_ambilight_power(True if power == "on" else False)


@ambilight.group("color", help="Manage ambilight color.")
def ambilight_color() -> None:
    pass


@ambilight_color.command("set")
@click.argument("color", type=(int, int, int), required=False)
@click.option("--left", "-l", type=(int, int, int), help="Color to set on the left side.")
@click.option("--top", "-t", type=(int, int, int), help="Color to set on the top side.")
@click.option("--right", "-r", type=(int, int, int), help="Color to set on the right side.")
@click.option("--bottom", "-b", type=(int, int, int), help="Color to set on the bottom side.")
@pass_tv_context
@handle_tv_errors
def ambilight_color_set(
    tv_ctx: TVContext,
    color: Optional[ColorArg],
    left: Optional[ColorArg],
    top: Optional[ColorArg],
    right: Optional[ColorArg],
    bottom: Optional[ColorArg],
) -> None:
    """Set ambilight color.

    If 'COLOR' positional argument is given, it will be set on all sides. If value for any of the
    sides is specified, it overrides the 'COLOR' value. Sides which aren't specified either through
    positional argument or through options, will keep their current color.

    Each color value should be specified in RGB as three integers in range 0-255.

    \b
    Examples:
        philipstv ambilight color set 255 0 0
        philipstv ambilight color set 255 0 0 --top 0 255 0
        philipstv ambilight color set --left 0 0 255 --right 0 0 255

    """
    if not any((color, left, top, right, bottom)):
        raise UsageError(
            "Missing argument 'COLOR' or one of: '--left', '--top', '--right', '--bottom'."
        )

    tv_ctx.remote.set_ambilight_color(
        color=AmbilightColor.from_tuple(color) if color else None,
        left=AmbilightColor.from_tuple(left) if left else None,
        top=AmbilightColor.from_tuple(top) if top else None,
        right=AmbilightColor.from_tuple(right) if right else None,
        bottom=AmbilightColor.from_tuple(bottom) if bottom else None,
    )


@cli.group("app", help="Manage applications.")
def app() -> None:
    pass


@app.command("list", help="List all available applications.")
@pass_tv_context
@handle_tv_errors
def app_list(tv_ctx: TVContext) -> None:
    click.echo("\n".join(tv_ctx.remote.get_applications()))


@app.command("launch", help="Launch an application.")
@click.argument("application", type=click.STRING)
@pass_tv_context
@handle_tv_errors
def app_launch(tv_ctx: TVContext, application: str) -> None:
    tv_ctx.remote.launch_application(application)



from typing import Any, Optional

from .model import PairingResponse


class PhilipsError(Exception):
    """Base exception for all exceptions raised by this package."""

    def __init__(self, message: str) -> None:
        super().__init__()
        self.message = message

    def __str__(self) -> str:
        return self.message


class PhilipsTVError(PhilipsError):
    """Raised if error occured during communication with the TV.

    Attributes:
        method: HTTP method of the request.
        url: URL of the request.
        status_code: HTTP status_code if response was received.

    """

    def __init__(self, method: str, url: str, status_code: Optional[int] = None) -> None:
        message = f"{method} request to {url} failed"
        if status_code:
            message += f" with status {status_code}"
        super().__init__(message)
        self.method = method
        self.url = url
        self.status_code = status_code


class PhilipsTVPairingError(PhilipsError):
    """Raised if pairing process failed.

    This could happen e.g. when requesting pairing during another pairing in progress, or if
    provided PIN is incorrect.

    Attributes:
        response: Response to the pairing request, containing error details.

    """

    def __init__(self, response: PairingResponse) -> None:
        super().__init__(f"Pairing error: {response.error_id} {response.error_text}")
        self.response = response


class PhilipsTVAPIError(PhilipsError):
    """Raised when something goes wrong with API communication."""


class PhilipsTVAPIUnauthorizedError(PhilipsTVAPIError):
    """Raised when trying to use the API without, or with invalid credentials.

    Attributes:
        method: HTTP method of the request.
        path: Path of the request.

    """

    def __init__(self, method: str, path: str) -> None:
        super().__init__(f"Unauthorized API access: {method} {path}")
        self.method = method
        self.path = path


class PhilipsTVAPIMalformedResponseError(PhilipsTVAPIError):
    """Raised when response received from the API is unparsable into known object.

    Attributes:
        method: HTTP method of the request.
        path: Path of the request.
        response: Raw response JSON returned by the API.

    """

    def __init__(self, method: str, path: str, response: Any) -> None:
        super().__init__(f"Malformed API response: {method} {path}")
        self.method = method
        self.path = path
        self.response = response


class PhilipsTVRemoteError(PhilipsError):
    """Raised if there was some incorrect interaction with :class:`PhilipsTVRemote`.

    This can happen e.g. when trying to launch application which is not installed. This exception
    does *NOT* come from the TV. It comes from the :class:`PhilipsTVRemote` class logic.
    """



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['philipstv', 'philipstv.model']

package_data = \
{'': ['*']}

install_requires = \
['pydantic>=1.9.0,<2.0.0', 'requests>=2.27.1,<3.0.0']

extras_require = \
{'cli': ['click>=8.0.3,<9.0.0', 'appdirs>=1.4.4,<2.0.0'],
 'docs': ['Sphinx>=4.4.0,<5.0.0',
          'furo>=2022.2.14,<2023.0.0',
          'enum-tools[sphinx]>=0.9.0,<0.10.0']}

entry_points = \
{'console_scripts': ['philipstv = philipstv.__main__:wrapped_cli']}

setup_kwargs = {
    'name': 'philipstv',
    'version': '0.4.0',
    'description': 'CLI and library to control Philips Android-powered TVs.',
    'long_description': "philipstv\n=========\n\n.. image:: https://github.com/bcyran/philipstv/workflows/CI/badge.svg?event=push\n   :target: https://github.com/bcyran/philipstv/actions?query=event%3Apush+branch%3Amaster+workflow%3ACI\n   :alt: CI\n\n.. image:: https://codecov.io/gh/bcyran/philipstv/branch/master/graph/badge.svg?token=ROJONX34RB\n   :target: https://codecov.io/gh/bcyran/philipstv\n   :alt: codecov\n\n.. image:: https://img.shields.io/badge/code%20style-black-000000.svg\n   :target: https://github.com/psf/black\n   :alt: Code style: black\n\n.. image:: https://img.shields.io/pypi/v/philipstv\n   :target: https://pypi.org/project/philipstv/\n   :alt: pypi\n\n.. image:: https://readthedocs.org/projects/philipstv/badge/?version=latest\n   :target: https://philipstv.readthedocs.io/en/latest/?badge=latest\n   :alt: Documentation status\n\n.. image:: https://img.shields.io/pypi/pyversions/philipstv\n   :target: https://pypi.org/project/philipstv/\n   :alt: versions\n\n.. image:: https://img.shields.io/github/license/bcyran/philipstv\n   :target: https://github.com/bcyran/philipstv/blob/master/LICENSE\n   :alt: license\n\n.. -begin-intro-\n\nPython package providing CLI and library for interacting with Philips Android-powered TVs.\n\nFeatures:\n\n- Get and set TV power state.\n- Get and set volume.\n- List and change TV channels.\n- Emulate pressing remote keys.\n- Get and set Ambilight power state.\n- Get and set Ambilight color.\n- List and launch applications.\n\nInstallation\n------------\n\nIf you plan to use the CLI:\n\n.. code-block:: console\n\n   $ pip install 'philipstv[cli]'\n\nIf you only need library for use in Python code:\n\n.. code-block:: console\n\n   $ pip install philipstv\n\n.. -end-intro-\n\nDocumentation\n-------------\nSee full documentation: `Read the Docs: philipstv <https://philipstv.readthedocs.io>`_.\n\nResources\n---------\n- `Fantastic unofficial API documentation <https://github.com/eslavnov/pylips/blob/master/docs/Home.md>`_ and `script <https://github.com/eslavnov/pylips>`_ by `@eslavnov <https://github.com/eslavnov>`_.\n- Philips `JointSpace API documentation <http://jointspace.sourceforge.net/projectdata/documentation/jasonApi/1/doc/API.html>`_.\n",
    'author': 'Bazyli Cyran',
    'author_email': 'bazyli@cyran.dev',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/bcyran/philipstv',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



from contextlib import contextmanager
from typing import Any, Iterator, Optional, Type, TypeVar

from philipstv.exceptions import (
    PhilipsTVAPIMalformedResponseError,
    PhilipsTVAPIUnauthorizedError,
    PhilipsTVError,
)

from .model import (
    AllChannels,
    AmbilightColors,
    AmbilightColorSettings,
    AmbilightMode,
    AmbilightPower,
    AmbilightTopology,
    APIObject,
    Applications,
    ApplicationShort,
    CurrentChannel,
    CurrentVolume,
    InputKey,
    PairingGrantPayload,
    PairingRequestPayload,
    PairingRequestResponse,
    PairingResponse,
    PowerState,
    SetChannel,
    ValidationError,
    Volume,
)
from .tv import PhilipsTV
from .types import Credentials

_T = TypeVar("_T", bound=APIObject)


@contextmanager
def _wrap_unauthorized_exceptions(path: str) -> Iterator[None]:
    try:
        yield
    except PhilipsTVError as exc:
        if exc.status_code == 401:
            raise PhilipsTVAPIUnauthorizedError(exc.method, path) from exc
        raise


@contextmanager
def _wrap_validation_exceptions(method: str, path: str, response: Any) -> Iterator[None]:
    try:
        yield
    except ValidationError as exc:
        raise PhilipsTVAPIMalformedResponseError(method, path, response) from exc


class PhilipsTVAPI:
    """Wrapper around Philips TV API.

    Each method mirrors a single `GET` or `POST` request to one of the available API endpoints.
    The original shape of request payloads and responses in preserved, but they're wrapped
    in `pydantic` models.

    """

    def __init__(self, tv: PhilipsTV) -> None:
        """
        Args:
            tv: Instance of the :class:`PhilipsTV` to which the API requestes will be sent.

        """
        self._tv = tv
        self.api_version = 6

    @property
    def host(self) -> str:
        """IP address of the underlying :class:`PhilipsTV` instance.

        Warning:
            This value is read-only.

        """
        return self._tv.host

    @property
    def auth(self) -> Optional[Credentials]:
        """Credentials used for authentication in the underlying :class:`PhilipsTV` instance.

        Hint:
            This value can be set and changed at any moment during :class:`PhilipsTVAPI` usage.

        """
        return self._tv.auth

    @auth.setter
    def auth(self, auth: Optional[Credentials]) -> None:
        self._tv.auth = auth

    def pair_request(self, payload: PairingRequestPayload) -> PairingRequestResponse:
        """Send request initiating the pairing process.

        Use :func:`pair_grant` to finalize the pairing using data from response to this request.

        Args:
            payload: Request payload.

        Returns:
            The TV response containing authentication details.

        """
        return self._api_post_model("pair/request", PairingRequestResponse, payload)

    def pair_grant(self, payload: PairingGrantPayload) -> PairingResponse:
        """Send request finalizing the pairing process.

        This requires some data from response to request initiating the pairing:
        :func:`pair_request`.

        Args:
            payload: Request payload.

        Returns:
            The TV response containing pairing confirmation or rejection.

        """
        return self._api_post_model("pair/grant", PairingResponse, payload)

    def get_powerstate(self) -> PowerState:
        """Send request to get the current power state."""
        return self._api_get_model("powerstate", PowerState)

    def set_powerstate(self, powerstate: PowerState) -> None:
        """Send request to set the power state."""
        self._api_post("powerstate", powerstate)

    def get_volume(self) -> CurrentVolume:
        """Send request to get the current volume, mute status, and volume limits."""
        return self._api_get_model("audio/volume", CurrentVolume)

    def set_volume(self, volume: Volume) -> None:
        """Send request to set the volume."""
        self._api_post("audio/volume", volume)

    def get_current_channel(self) -> CurrentChannel:
        """Send request to get current TV app activity."""
        return self._api_get_model("activities/tv", CurrentChannel)

    def get_all_channels(self) -> AllChannels:
        """Send request to get all available channels."""
        return self._api_get_model("channeldb/tv/channelLists/all", AllChannels)

    def set_channel(self, channel: SetChannel) -> None:
        """Send request to set the channel."""
        self._api_post("activities/tv", channel)

    def input_key(self, key: InputKey) -> None:
        """Send request to simulate pressing key on the remote."""
        self._api_post("input/key", key)

    def get_ambilight_power(self) -> AmbilightPower:
        """Send request to get Ambilight power state."""
        return self._api_get_model("ambilight/power", AmbilightPower)

    def set_ambilight_power(self, power: AmbilightPower) -> None:
        """Send request to set Ambilight power state."""
        self._api_post("ambilight/power", power)

    def get_ambilight_topology(self) -> AmbilightTopology:
        """Send request to get Ambilight topology."""
        return self._api_get_model("ambilight/topology", AmbilightTopology)

    def get_ambilight_mode(self) -> AmbilightMode:
        """Send request to get current Ambilight mode."""
        return self._api_get_model("ambilight/mode", AmbilightMode)

    def set_ambilight_mode(self, mode: AmbilightMode) -> None:
        """Send request to set ambilight mode."""
        self._api_post("ambilight/mode", mode)

    def get_ambilight_measured(self) -> AmbilightColors:
        """Send request to get measured color values from Ambilight system.

        Those are the colors taken directly from the displayed image, before processing by
        internal Ambilight algorithm.
        """
        return self._api_get_model("ambilight/measured", AmbilightColors)

    def get_ambilight_processed(self) -> AmbilightColors:
        """Send request to get processed color values from Ambilight system.

        Those are the measured values after processing by the internal Ambilight algorithm, based on
        the currently active Ambilight style.
        """
        return self._api_get_model("ambilight/processed", AmbilightColors)

    def get_ambilight_cached(self) -> AmbilightColors:
        """Send request to get cached color values from Ambilight system.

        Those are the values that were previously set by some user via API.
        """
        return self._api_get_model("ambilight/cached", AmbilightColors)

    def set_ambilight_cached(self, colors: AmbilightColorSettings) -> None:
        """Send request to set cached color values in Ambilight system.

        If those values are set, they override other Ambilight settings. This effectively just sets
        the Ambilight color.

        Note:
            The actual TV API allows sending the colors in one of the following ways:
                1. All pixels on all layers to the same color.
                2. All pixels on a single layer to the same color.
                3. Specific colors for pixels on chosen sides (left, top, etc.) of a single layer.
                4. Specific color for any pixel addressed by layer, side and index.

            This implementation allows only options 1 and 4. I have no idea what are the "layers".
            The TVs I have access to have only one. I wasn't able to find any information about it
            so I decided to ignore it. Option 3 could actually be useful, I'll consider implementing
            it, but right now the same effect can be achieved using option 4.

        All pixels which color's are not set in a given request (e.g. setting a single pixel's
        color), will keep their color.

        References:
            `Philips JointSpace API Documentation
            <http://jointspace.sourceforge.net/projectdata/documentation/jasonApi/1/doc/API-Method-ambilight-cached-POST.html>`_

        """
        self._api_post("ambilight/cached", colors)

    def get_applications(self) -> Applications:
        """Send request to get the list of installed applications."""
        return self._api_get_model("applications", Applications)

    def launch_application(self, application: ApplicationShort) -> None:
        """Send request to launch an application activity.

        Use values returned by :func:`get_applications`.

        Note:
            "Activity" means that this can go directly to a specific screen in the application.
            Unfortunately, I don't know where to get valid "activities" for specific applications.

        """
        self._api_post("activities/launch", application)

    def _api_post_model(
        self, path: str, resp_model: Type[_T], payload: Optional[APIObject] = None
    ) -> _T:
        raw_response = self._api_post(path, payload)
        with _wrap_validation_exceptions("POST", path, raw_response):
            return resp_model.parse(raw_response)

    def _api_get_model(self, path: str, response_model: Type[_T]) -> _T:
        raw_response = self._api_get(path)
        with _wrap_validation_exceptions("GET", path, raw_response):
            return response_model.parse(raw_response)

    def _api_post(self, path: str, payload: Optional[APIObject] = None) -> Any:
        with _wrap_unauthorized_exceptions(path):
            return self._tv.post(self._api_path(path), payload.dump() if payload else None)

    def _api_get(self, path: str) -> Any:
        with _wrap_unauthorized_exceptions(path):
            return self._tv.get(self._api_path(path))

    def _api_path(self, path: str) -> str:
        return f"{self.api_version}/{path}"



import logging
from contextlib import contextmanager
from typing import Any, Iterator, Optional
from urllib.parse import urljoin

import urllib3
from requests import RequestException, Session
from requests.auth import HTTPDigestAuth
from urllib3.exceptions import InsecureRequestWarning

from .exceptions import PhilipsTVError
from .types import Credentials

urllib3.disable_warnings(InsecureRequestWarning)  # type: ignore


_LOGGER = logging.getLogger(__name__)


@contextmanager
def _wrap_http_exceptions() -> Iterator[None]:
    try:
        yield
    except RequestException as exc:
        status_code = exc.response.status_code if exc.response is not None else None
        raise PhilipsTVError(exc.request.method, exc.request.url, status_code) from exc


class PhilipsTV:
    """Lowest level interface with the TV.

    Used to send raw (without using model objects) `GET` and `POST` requests to any API path.

    This handles authentication, exception handling, logging and URL building so higher layers don't
    have to worry about this.
    """

    def __init__(self, host: str, port: int = 1926, auth: Optional[Credentials] = None) -> None:
        """
        Args:
            host: TV IP address to connect to.
            port: TV port to connect to.
            auth: Authentication credentials tuple.

        """
        self.host = host
        self.port = port
        self.url = f"https://{self.host}:{self.port}"

        self._auth: Optional[Credentials] = None
        self._session = self._create_session()
        self.auth = auth

    @property
    def auth(self) -> Optional[Credentials]:
        """Credentials used for authentication.

        Hint:
            This value can be set and changed at any moment during :class:`PhilipsTV` usage.

        """
        return self._auth

    @auth.setter
    def auth(self, value: Optional[Credentials]) -> None:
        self._session.auth = HTTPDigestAuth(*value) if value else None
        self._auth = value

    def post(self, path: str, payload: Any = None) -> Any:
        """Send `POST` request.

        Args:
            path: The path to send the request to.
            payload: Request payload to send.

        Returns:
            The TV's JSON response body or ``None``.

        """
        _LOGGER.debug("Request: POST %s %s", path, payload)
        with _wrap_http_exceptions():
            response = self._session.post(urljoin(self.url, path), json=payload)
            response.raise_for_status()
        response_body = response.json() if response.content else None
        _LOGGER.debug("Response: %s %s", response.status_code, response_body)
        return response_body

    def get(self, path: str) -> Any:
        """Send `GET` request.

        Args:
            path: The path to send the request to.

        Returns:
            The TV's JSON response body or ``None``.

        """
        _LOGGER.debug("Request: GET %s", path)
        with _wrap_http_exceptions():
            response = self._session.get(urljoin(self.url, path))
            response.raise_for_status()
        response_body = response.json() if response.content else None
        _LOGGER.debug("Response: %s %s", response.status_code, response_body)
        return response_body

    @staticmethod
    def _create_session() -> Session:
        session = Session()
        session.verify = False
        return session



from typing import List

from pydantic import Field

from .base import APIObject


class ChannelID(APIObject):
    """Model of a channel change request."""

    ccid: int
    """ID of the channel."""


class ChannelShort(ChannelID):
    """Model of a channel definition in current channel response."""

    preset: str
    """Number by which the channel is available on the TV."""
    name: str
    """Name of the channel."""


class Channel(ChannelShort):
    """Model of a channel definition in channels list response."""

    onid: int
    """Original Network ID."""
    tsid: int
    """Transport Stream ID."""
    sid: int
    """Service ID."""
    service_type: str = Field(alias="serviceType")
    """Service type."""
    type: str
    """Channel type."""
    logo_version: int = Field(alias="logoVersion")
    """Channel logo version."""


class ChannelListID(APIObject):
    """Model of a channels list definition in channel change request."""

    id: str = "allcab"
    """ID of the channel list."""


class ChannelList(ChannelListID):
    """Model of a channels list definition in channels list response."""

    version: str
    """Channel list version."""


class CurrentChannel(APIObject):
    """Model of a current channel response."""

    channel: ChannelShort
    """Current channel definition."""
    channel_list: ChannelList = Field(alias="channelList")
    """Current channel list definition."""


class SetChannel(APIObject):
    """Model of a channel change request."""

    channel: ChannelID
    """Definition of a channel to change to."""
    channel_list: ChannelListID = Field(default_factory=ChannelListID, alias="channelList")
    """Definition of a list this channel is on."""


class AllChannels(APIObject):
    """Model of all channels list response."""

    version: int
    """List version."""
    id: str
    """List ID."""
    list_type: str = Field(alias="listType")
    """List type."""
    medium: str
    """TV reception medium."""
    operator: str
    """TC operator name."""
    install_country: str = Field(alias="installCountry")
    """Country the TV service is installed in."""
    channel: List[Channel] = Field(alias="Channel")
    """List of channel definitions."""



import hmac
import random
import string
from base64 import b64encode
from hashlib import sha256


def create_device_id() -> str:
    return "".join(
        random.choice(string.ascii_uppercase + string.ascii_lowercase + string.digits)
        for _ in range(16)
    )


def create_signature(secret: bytes, message: bytes) -> bytes:
    return b64encode(hmac.new(secret, message, sha256).hexdigest().encode())
